__title__ = 'campy'
__description__ = "Stanford's Graphical Libraries in Python"
__url__ = 'https://campy.sredmond.io/'
__license__ = 'MIT'

__version__ = '0.0.1.dev11'
__build__ = 0x000001
__status__ = 'Alpha'

__author__ = 'Sam Redmond'
__maintainer__ = 'Sam Redmond'
__email__ = 'sredmond@stanford.edu'
__copyright__ = '(c) 2016-2019 by Sam Redmond'
__credits__ = ['Sam Redmond', 'Alex Valderrama', 'Marty Stepp', 'Eric Roberts']

__snake__ = '\u2728 \U0001F40D \u2728'
