"""Welcome to `campy.stanford`.

See:
    https://www.stanforddaily.com/2011/04/26/one-time-at-camp-stanford/
"""
import sys
if sys.stdout.isatty():
    print('Importing `campy.stanford`...')
    print('Import failed.')
    print('Importing `camp.stanford` instead...')
    print('☀️🌲🏕️🌲☀️')
